# -*- coding: utf-8 -*-
"""Functions to load datasets included in sktime."""

__all__ = [
    "load_airline",
    "load_arrow_head",
    "load_gunpoint",
    "load_basic_motions",
    "load_osuleaf",
    "load_italy_power_demand",
    "load_japanese_vowels",
    "load_longley",
    "load_lynx",
    "load_shampoo_sales",
    "load_UCR_UEA_dataset",
    "load_unit_test",
    "load_uschange",
    "load_PBS_dataset",
    "load_japanese_vowels",
    "load_acsf1",
]

from sktime.datasets._data_io import load_airline
from sktime.datasets._data_io import load_gunpoint
from sktime.datasets._data_io import load_arrow_head
from sktime.datasets._data_io import load_basic_motions
from sktime.datasets._data_io import load_osuleaf
from sktime.datasets._data_io import load_italy_power_demand
from sktime.datasets._data_io import load_japanese_vowels
from sktime.datasets._data_io import load_longley
from sktime.datasets._data_io import load_lynx
from sktime.datasets._data_io import load_shampoo_sales
from sktime.datasets._data_io import load_UCR_UEA_dataset
from sktime.datasets._data_io import load_unit_test
from sktime.datasets._data_io import load_uschange
from sktime.datasets._data_io import load_PBS_dataset
from sktime.datasets._data_io import load_acsf1
